import math
from typing import Tuple, List

from shapely.geometry import Polygon, Point, MultiPoint, LineString, LinearRing, MultiPolygon, MultiLineString, \
    GeometryCollection
from shapely.geometry.base import BaseGeometry


def interpolate_coords_by_len(coord1: Tuple[float, float],
                              coord2: Tuple[float, float],
                              gap: float,
                              result_include_coord2: bool = True) -> List[Tuple[float, float]]:
    vector = (coord2[0] - coord1[0], coord2[1] - coord1[1])
    vec_len = math.sqrt(vector[0] ** 2 + vector[1] ** 2)
    unit_vector = (vector[0] / vec_len, vector[1] / vec_len)

    result = [coord1]
    for i in range(1, round(vec_len / gap)):
        result.append((coord1[0] + unit_vector[0] * i * gap, coord1[1] + unit_vector[1] * i * gap))
    if result_include_coord2:
        result.append(coord2)
    return result


def interpolate(geometry: BaseGeometry, gap: float, simplify_distance: float = 1e-6) -> BaseGeometry:
    def interpolate_coords(coords):
        new_coords: List[Tuple[float, float]] = []
        for i in range(len(coords) - 2):
            new_coords.extend(interpolate_coords_by_len(coords[i], coords[i + 1], gap, result_include_coord2=False))
        new_coords.extend(interpolate_coords_by_len(coords[-2], coords[-1], gap, result_include_coord2=True))
        return new_coords

    if isinstance(geometry, (Point, MultiPoint)):
        return geometry
    elif isinstance(geometry, (Polygon, LineString, LinearRing)):
        geometry_simplified = geometry.simplify(simplify_distance)
        if isinstance(geometry, Polygon):
            exterior_coords = list(geometry_simplified.exterior.coords)
            interior_coords_list = [list(interior.coords) for interior in geometry_simplified.interiors]
            interpolated_exterior = interpolate_coords(exterior_coords)
            interpolated_interiors = [interpolate_coords(interior_coords) for interior_coords in interior_coords_list]
            return Polygon(shell=interpolated_exterior, holes=interpolated_interiors)
        else:  # LineString or LinearRing
            coords = list(geometry_simplified.coords)
            return type(geometry)(interpolate_coords(coords))
    elif isinstance(geometry, (MultiPolygon, MultiLineString, GeometryCollection)):
        geoms = list(geometry)
        interpolated_geoms = [interpolate(geom, gap) for geom in geoms]
        return type(geometry)(interpolated_geoms)

    return geometry  # return origin geometry if not match any type



from typing import List

import numpy as np
from scipy.spatial import Delaunay
from shapely.geometry import Polygon

from shapely_ext.interpolate import interpolate


class TriMesher:
    def __init__(self, interpolate_distance: float):
        self._interpolate_distance = interpolate_distance

    def mesh(self, polygon: Polygon) -> List[Polygon]:
        if not isinstance(polygon, Polygon):
            raise NotImplementedError("only polygon can be meshed")
        interpolated_polygon = interpolate(polygon, gap=self._interpolate_distance)
        exterior_coords = np.array(list(interpolated_polygon.exterior.coords))
        delaunay = Delaunay(exterior_coords)
        mesh_faces: List[Polygon] = list(
            map(lambda tri_coords: Polygon(tri_coords), exterior_coords[delaunay.simplices]))
        return mesh_faces



import math
from typing import Tuple, List

from shapely.geometry import LineString, Polygon


class AngleMeasurer:
    def __init__(self, simplify_distance: float = 1e-6):
        self._simplify_distance = simplify_distance

    def get_straight_line_angle(self, line: LineString, in_degree: bool = True) -> float:
        """
        return angle in range [-90, 90]
        """
        if not line.is_valid:
            raise ValueError("input line is not a valid lineString")
        coords = list(line.coords)
        endpoint1, endpoint2 = coords[0], coords[-1]

        angle_in_radian = (math.atan2(endpoint2[1] - endpoint1[1], endpoint2[0] - endpoint1[0])) % math.pi
        if angle_in_radian > math.pi / 2:
            angle_in_radian -= math.pi
        if in_degree:
            return math.degrees(angle_in_radian)
        return angle_in_radian

    def get_angle_by_coords(self, coord1: Tuple[float, float], coord2: Tuple[float, float],
                            in_degree: bool = True) -> float:
        if coord1 == coord2:
            raise ValueError("coord1 and coord2 should not be equal")
        line = LineString([coord1, coord2])
        return self.get_straight_line_angle(line, in_degree=in_degree)

    def get_polygon_angle_by_bounding_box(self, polygon: Polygon, in_degree: bool = True) -> float:
        """
        计算polygon的最小外接矩形的长边的角度, return angle in range [-90, 90]
        """
        rotated_box_coords = list(polygon.minimum_rotated_rectangle.simplify(self._simplify_distance).exterior.coords)
        edges = [(rotated_box_coords[i], rotated_box_coords[i + 1]) for i in range(2)]
        longest_edge = max(edges,
                           key=lambda coords: (coords[0][0] - coords[1][0]) ** 2 + (coords[0][1] - coords[1][1]) ** 2)
        return self.get_angle_by_coords(*longest_edge, in_degree=in_degree)



from unittest import TestCase

from shapely.geometry import box, Polygon, Point

from shapely_ext.mesh import TriMesher


class Test(TestCase):
    def setUp(self) -> None:
        self.mesher_11 = TriMesher(interpolate_distance=11)
        self.mesher_5 = TriMesher(interpolate_distance=5)

    def test_generate_tri_mesh_of_polygon(self):
        rect = box(0, 0, 10, 10)
        faces = self.mesher_11.mesh(rect)  # non interpolation
        self.assertTrue(all(isinstance(face, Polygon) for face in faces))
        self.assertTrue(2, len(faces))

        faces = self.mesher_5.mesh(rect)
        self.assertTrue(all(isinstance(face, Polygon) for face in faces))
        self.assertTrue(8, len(faces))

    def test_generate_tri_mesh_of_non_polygon(self):
        point = Point(0, 0)
        with self.assertRaises(NotImplementedError):
            self.mesher_11.mesh(point)



import math
from typing import Sequence, TypeVar

from shapely.affinity import translate
from shapely.geometry.base import BaseGeometry

Num = TypeVar('Num', float, int)


class Vector2D:
    def __init__(self, x, y):
        if not isinstance(x, (float, int)) or not isinstance(y, (float, int)):
            raise TypeError("x, y should be number")
        self.x = x
        self.y = y

    @staticmethod
    def is_valid_2d_coordinate(coord: Sequence[Num]):
        return (isinstance(coord, Sequence)
                and len(coord) >= 2
                and isinstance(coord[0], (int, float))
                and isinstance(coord[1], (int, float)))

    @classmethod
    def from_coordinate(cls, coord: Sequence[Num]):
        if not cls.is_valid_2d_coordinate(coord):
            raise ValueError(f"{coord} is not valid coordinates")
        return cls(x=coord[0], y=coord[1])

    @classmethod
    def from_coordinates(cls, from_coord: Sequence[Num], to_coord: Sequence[Num]):
        if not cls.is_valid_2d_coordinate(from_coord):
            raise ValueError(f"{from_coord} is not valid from_coordinates")
        if not cls.is_valid_2d_coordinate(to_coord):
            raise ValueError(f"{to_coord} is not valid from_coordinates")
        return cls(x=to_coord[0] - from_coord[0], y=to_coord[1] - from_coord[1])

    @property
    def length(self):
        return math.sqrt(self.x ** 2 + self.y ** 2)

    def __eq__(self, other):
        return self.x == other.x and self.y == other.y

    @classmethod
    def raise_if_not_vector(cls, possible_vector):
        if not isinstance(possible_vector, cls):
            raise TypeError(f"{possible_vector} is of type {type(possible_vector)}, expect Vector2D")

    def plus(self, vector):
        self.raise_if_not_vector(vector)
        return Vector2D(self.x + vector.x, self.y + vector.y)

    def dot(self, vector) -> Num:
        self.raise_if_not_vector(vector)
        return self.x * vector.x + self.y * vector.y

    def angle_ccw_rotating_to(self, vector_2d, in_degree: bool = True):
        """
        in range of [0, 360] degree
        """
        dot_prod = self.dot(vector_2d)
        cos_val = dot_prod / self.length / vector_2d.length
        angle_in_radian = math.acos(cos_val)
        cross_prod = self.x * vector_2d.y - self.y * vector_2d.x
        if cross_prod < 0:  # then vector_2d is on the cw side of self vector
            angle_in_radian = math.pi * 2 - angle_in_radian
        if in_degree:
            return math.degrees(angle_in_radian)
        return angle_in_radian

    def angle_to(self, vector_2d, in_degree: bool = True):
        """
        in range of [0, 180]
        """
        ccw_rotating_angle_degree = self.angle_ccw_rotating_to(vector_2d)
        if ccw_rotating_angle_degree > 180:
            ccw_rotating_angle_degree = 360 - ccw_rotating_angle_degree
        if not in_degree:
            return math.radians(ccw_rotating_angle_degree)
        return ccw_rotating_angle_degree

    def apply(self, geom: BaseGeometry) -> BaseGeometry:
        return translate(geom, xoff=self.x, yoff=self.y)



from unittest import TestCase

from shapely.geometry import box

from shapely_ext.func import group


class TestGroup(TestCase):
    def test_group_num(self):
        items = [0, 0.1, 0.2, 0.3, 2, 2.1, 2.12, 10, 10.1]
        groups = group(items, grouping_func=lambda x1, x2: abs(x1 - x2) < 0.15)
        self.assertEqual(3, len(groups))
        groups.sort(key=len)
        self.assertListEqual([10, 10.1], sorted(groups[0]))
        self.assertListEqual([2, 2.1, 2.12], sorted(groups[1]))
        self.assertListEqual([0, 0.1, 0.2, 0.3], sorted(groups[2]))

    def test_group_polygon(self):
        boxes = [
            box(0, 0, 1, 1),
            box(1.1, 1, 2, 2),
            box(-10, 0, -9, 1),
        ]
        groups = group(boxes, grouping_func=lambda b1, b2: b1.distance(b2) < 0.5)
        self.assertEqual(2, len(groups))
        groups.sort(key=len)
        self.assertListEqual([boxes[2]], groups[0])
        self.assertListEqual(boxes[:2], groups[1])



from itertools import chain
from typing import List

from shapely.geometry import Polygon, LineString, LinearRing, MultiPolygon, MultiLineString, GeometryCollection
from shapely.geometry.base import BaseGeometry

from shapely_ext.geometry.vector_2d import Vector2D


class Decomposer:
    def __init__(self, min_corner_angle_degree: float = 0, simplify_distance: float = 1e-6):
        self._min_corner_angle_degree = min_corner_angle_degree
        self._simplify_distance = simplify_distance

    def decompose(self, geometry: BaseGeometry) -> List[LineString]:
        geometry = geometry.simplify(self._simplify_distance)
        if isinstance(geometry, Polygon):
            return self.decompose_polygon(polygon=geometry)
        elif isinstance(geometry, MultiPolygon):
            lines_list: List[List[LineString]] = [
                self.decompose_polygon(polygon=sub_polygon)
                for sub_polygon in list(geometry)]
            return list(chain.from_iterable(lines_list))
        elif isinstance(geometry, LinearRing):
            return self.decompose_linearRing(ring=geometry)
        elif isinstance(geometry, LineString):
            return self.decompose_lineString(lineString=geometry)
        elif isinstance(geometry, MultiLineString):
            lines_list: List[List[LineString]] = [
                self.decompose_lineString(lineString=sub_lineString)
                for sub_lineString in list(geometry)]
            return list(chain.from_iterable(lines_list))
        elif isinstance(geometry, GeometryCollection):
            lines_list: List[List[LineString]] = [
                self.decompose(geometry=sub_geometry) for sub_geometry in list(geometry)]
            return list(chain.from_iterable(lines_list))
        else:
            return []

    def decompose_lineString(self, lineString: LineString):
        coords = list(lineString.coords)
        lines: List[LineString] = []
        cur_line_coords = []
        for i in range(len(coords)):
            cur_line_coords.append(coords[i])
            if i - 1 < 0 or i + 1 >= len(coords):
                continue

            if self._is_corner(coords[i], coords[i - 1], coords[i + 1],
                               min_corner_angle_degree=self._min_corner_angle_degree):
                lines.append(LineString(cur_line_coords))
                cur_line_coords = [coords[i]]
        if cur_line_coords:
            lines.append(LineString(cur_line_coords))
        return lines

    def decompose_polygon(self, polygon: Polygon) -> List[LineString]:
        exterior_lines: List[LineString] = self.decompose_linearRing(ring=polygon.exterior)
        interior_lines_list: List[List[LineString]] = [
            self.decompose_linearRing(ring=interior)
            for interior in polygon.interiors
        ]
        return exterior_lines + list(chain.from_iterable(interior_lines_list))

    def decompose_linearRing(self, ring: LinearRing) -> List[LineString]:
        coords = list(ring.coords)[:-1]
        corner_i = self._find_first_corner_index(coords)
        lines: List[LineString] = []
        cur_line = [coords[corner_i]]
        for j in range(1, len(coords) + 1):
            cur_i = (corner_i + j) % len(coords)
            cur_line.append(coords[cur_i])
            if self._is_corner(coords[cur_i], coords[cur_i - 1],
                               coords[(cur_i + 1) % len(coords)],
                               min_corner_angle_degree=self._min_corner_angle_degree):
                lines.append(LineString(cur_line))
                cur_line = [coords[cur_i]]
        return lines

    def _find_first_corner_index(self, coords):
        corner_i = None
        for i in range(len(coords)):
            last_last_coord = coords[i - 2]
            last_coord = coords[i - 1]
            cur_coord = coords[i]
            if self._is_corner(last_coord,
                               last_last_coord,
                               cur_coord,
                               min_corner_angle_degree=self._min_corner_angle_degree):
                corner_i = i - 1
                break
        if corner_i is None:
            raise ValueError("Polygon has no corner")
        return corner_i

    @staticmethod
    def _is_corner(cur_coord, prev_coord, next_coord, min_corner_angle_degree: float) -> bool:
        prev_vec = Vector2D.from_coordinates(prev_coord, cur_coord)
        next_vec = Vector2D.from_coordinates(cur_coord, next_coord)
        angle_in_degree = prev_vec.angle_to(next_vec)
        return angle_in_degree >= min_corner_angle_degree



import math
from unittest import TestCase

from shapely.geometry import LineString, Polygon

from shapely_ext.angle import AngleMeasurer


class Test(TestCase):
    def setUp(self) -> None:
        self.angle_measurer = AngleMeasurer()

    def test_get_straight_line_angle(self):
        line1 = LineString([(0, 0), (1, 1)])
        self.assertAlmostEqual(45, self.angle_measurer.get_straight_line_angle(line1, in_degree=True), delta=1e-6)

        line2 = LineString([(0, 0), (1, 0)])
        self.assertAlmostEqual(0, self.angle_measurer.get_straight_line_angle(line2, in_degree=True), delta=1e-6)

    def test_get_angle_by_coords(self):
        self.assertAlmostEqual(45, self.angle_measurer.get_angle_by_coords((0, 0), (1, 1), in_degree=True), delta=1e-6)
        self.assertAlmostEqual(0, self.angle_measurer.get_angle_by_coords((0, 0), (1, 0), in_degree=True), delta=1e-6)

    def test_get_polygon_angle_in_degree(self):
        polygon1 = Polygon([(0, 0), (100, 0), (100, 10), (0, 10)])
        polygon2 = Polygon([(0, 0), (10, 0), (10, 100), (0, 100)])
        polygon3 = Polygon([(1, 0), (11, 10), (10, 11), (0, 1)])
        polygon4 = Polygon([(1, 0), (11, -10), (10, -11), (0, -1)])

        self.assertEqual(0, self.angle_measurer.get_polygon_angle_by_bounding_box(polygon1))
        self.assertEqual(90, self.angle_measurer.get_polygon_angle_by_bounding_box(polygon2, in_degree=True))
        self.assertAlmostEqual(math.pi / 2,
                               self.angle_measurer.get_polygon_angle_by_bounding_box(polygon2, in_degree=False),
                               delta=1e-6)
        self.assertAlmostEqual(45, self.angle_measurer.get_polygon_angle_by_bounding_box(polygon3, in_degree=True),
                               delta=1e-6)
        self.assertAlmostEqual(45, self.angle_measurer.get_polygon_angle_by_bounding_box(polygon3, in_degree=True))
        self.assertAlmostEqual(math.pi / 4,
                               self.angle_measurer.get_polygon_angle_by_bounding_box(polygon3, in_degree=False))
        self.assertAlmostEqual(-45, self.angle_measurer.get_polygon_angle_by_bounding_box(polygon4), delta=1e-6)
        self.assertAlmostEqual(-math.pi / 4,
                               self.angle_measurer.get_polygon_angle_by_bounding_box(polygon4, in_degree=False),
                               delta=1e-6)



import math
from typing import List
from unittest import TestCase

from shapely.geometry import box, LineString, Polygon, Point
from shapely.geometry.base import BaseGeometry
from shapely.ops import unary_union

from shapely_ext.decompose import Decomposer
from test.constant import MATH_EPS


class TestDecompose(TestCase):
    def setUp(self):
        self.decomposer1 = Decomposer()
        self.decomposer_angle_tol = Decomposer(min_corner_angle_degree=15)

    def test_decompose_polygon(self):
        polygon1 = box(0, 0, 10, 5)
        lines1 = self.decomposer1.decompose(geometry=polygon1)
        self.assertEqual(4, len(lines1))
        self._test_objects_are_all_of_type(lines1, LineString)
        self._test_objects_are_of_lens(lines1, [10, 10, 5, 5])

        polygon2 = Polygon([(0, 0), (1, 1), (0, 1)])
        lines2 = self.decomposer1.decompose(geometry=polygon2)
        self.assertEqual(3, len(lines2))
        self._test_objects_are_all_of_type(lines2, LineString)
        self._test_objects_are_of_lens(lines2, [1, 1, 2 ** 0.5])

        circle = Point(0.5, 0.5).buffer(0.5)
        polygon3 = unary_union([circle, box(0.5, 0, 1.5, 1)])
        line3 = self.decomposer_angle_tol.decompose(polygon3)
        self.assertEqual(2, len(line3))
        self._test_objects_are_all_of_type(line3, LineString)
        self._test_objects_are_of_lens(line3, [1, math.pi / 2 + 2])

    def _test_objects_are_all_of_type(self, geoms: List[BaseGeometry], type_class):
        return self.assertTrue(all(isinstance(g, type_class) for g in geoms))

    def _test_objects_are_of_lens(self, geoms: List[BaseGeometry], lens: List[float], delta: float = MATH_EPS):
        geom_lens = sorted(map(lambda g: g.length, geoms))
        lens.sort()
        for geom_len, expected_len in zip(geom_lens, lens):
            self.assertAlmostEqual(expected_len, geom_len, delta=delta)



from unittest import TestCase

from shapely.geometry import box, Polygon, LineString, MultiPoint, Point, GeometryCollection

from shapely_ext.interpolate import interpolate


class TestInterpolate(TestCase):
    def test_interpolate(self):
        rect = box(0, 0, 10, 10)
        interpolated_rect = interpolate(rect, 5)
        self.assertTrue(type(rect) is type(interpolated_rect))
        self.assertTrue(len(list(rect.exterior.coords)) < len(list(interpolated_rect.exterior.coords)))
        self.assertEqual(9, len(list(interpolated_rect.exterior.coords)))

        polygon = Polygon([(0, 0), (10, 0), (10, 10), (0, 10)], [[(4, 4), (4, 6), (6, 6), (6, 4)]])
        interpolated_polygon = interpolate(polygon, 1)
        self.assertTrue(type(polygon) is type(interpolated_polygon))
        self.assertTrue(len(list(polygon.exterior.coords)) < len(list(interpolated_polygon.exterior.coords)))
        self.assertTrue(
            len(list(polygon.interiors[0].coords)) < len(list(interpolated_polygon.interiors[0].coords)))
        self.assertEqual(41, len(list(interpolated_polygon.exterior.coords)))
        self.assertEqual(9, len(list(interpolated_polygon.interiors[0].coords)))

        line = LineString([(0, 0), (10, 0)])
        interpolated_line = interpolate(line, 1)
        self.assertTrue(type(line) is type(interpolated_line))
        self.assertTrue(len(line.coords) < len(interpolated_line.coords))
        self.assertEqual(11, len(interpolated_line.coords))

        point = Point(0, 0)
        interpolated_point = interpolate(point, 1)
        self.assertTrue(type(point) is type(interpolated_point))
        self.assertEqual(point, interpolated_point)

        mp = MultiPoint([point, point])
        interpolated_mp = interpolate(mp, 1)
        self.assertTrue(type(mp) is type(interpolated_mp))
        self.assertEqual(mp, interpolated_mp)

        collection = GeometryCollection([mp, point, line, polygon])
        interpolated_collection = interpolate(collection, 1)
        self.assertTrue(type(collection) is type(interpolated_collection))
        self.assertEqual(len(collection), len(list(interpolated_collection)))
        processed_polygon = list(filter(lambda geom: isinstance(geom, Polygon), list(interpolated_collection)))[0]
        self.assertEqual(41, len(list(processed_polygon.exterior.coords)))
        self.assertEqual(9, len(list(processed_polygon.interiors[0].coords)))
        processed_line = list(filter(lambda geom: isinstance(geom, LineString), list(interpolated_collection)))[0]
        self.assertEqual(11, len(processed_line.coords))



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="shapely_ext",
    version="0.1.3",
    author="pd",
    author_email="pyeprog@foxmail.com",
    description="a shapely extension package providing magic function",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/pyeprog/shapely_ext",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
    install_requires=[
        "Shapely>=1.7.0",
        "scipy>=1.4.1",
        "numpy>=1.18.4"
    ]
)


from typing import List, Any, Callable, Set


def group(items: List[Any],
          grouping_func: Callable[[Any, Any], bool]) -> List[List[Any]]:
    result: List[List[Any]] = []

    ungrouped_index: Set[int] = set(range(len(items)))
    seen: Set[int] = set()
    while len(ungrouped_index) > 0:
        cur_idx = ungrouped_index.pop()
        cur_group: List[Any] = []
        group_candidate = {cur_idx}
        while len(group_candidate) > 0:
            cand_idx = group_candidate.pop()
            seen.add(cand_idx)
            cur_group.append(items[cand_idx])
            idx_nearby_of_cand = set(
                idx for idx in ungrouped_index if grouping_func(items[idx], items[cand_idx]) and idx not in seen)
            ungrouped_index.difference_update(idx_nearby_of_cand)
            group_candidate.update(idx_nearby_of_cand)

        result.append(cur_group)
    return result
